/************************************************************
* $Id: bot_ctrl.c,v 1.6 2014/11/21 11:07:14 mechanoid Exp mechanoid $ 
*       
*       OS : GNU/Linux 3.14.19 
* COMPILER : GNU project C and C++ Compiler  v. 4.9.1
*
*   AUTHOR : Evgeny S. Borisov
*
*  Glushkov Institute of Cybernetics
*  National Academy of Sciences of Ukraine
* 
*    http://www.mechanoid.kiev.ua
*  e-mail : cv@mechanoid.kiev.ua
* 
************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "frame.h"
#include "port_ctrl.h"
#include "bot_ctrl.h"
#include "commands.h"

// количество отброшеных кадров после комманды на движение камеры
#define CHECK_SKIP_FRAMES 4 

// количество отброшеных кадров перед реакцией на потерю объекта
#define FRAMES_LOST_OBJ 5

#define CMD_LEN 64


int bot_obj_found();
int bot_obj_lost();
int bot_obj_pathfinder();





int lost_obj=FRAMES_LOST_OBJ ; // счетчик отброшеных кадров при потере объекта

// смещёние центра объекта относительно центра картинки
int obj_shift_x;
int obj_shift_y;
int obj_r; // радиус области с объектом

char cmd[CMD_LEN]; // буфер команд

int barrier=0; // флаг "обнаружено препятствие"

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
int bot_init(){
   if( frame_init() ) { return 2; }
   if( port_ctrl_init() ) { return 4; }
	  
   return 0;
}
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// завершение работы
void bot_done(){
   // остановка моторов
   snprintf(cmd,CMD_LEN,"%c",CMD_MOT_STOP);
   port_ctrl_cmd(cmd,strnlen(cmd,CMD_LEN)); 

   frame_done();
   port_ctrl_done();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
int bot_run() {
   // char buf[1024];
   
   for(;;) {
	  // получить кадр 
	  if(frame_proc(&obj_shift_x,&obj_shift_y,&obj_r)) { return 3 ; }

	  if(obj_r) {
	     bot_obj_found(); // найден объект
	  } else {
	     bot_obj_lost(); // объект в кадре не обнаружен
	  }

	  /*
	  // проверяем сообщения от реактивного уровня 
	  if( port_ctrl_answer(buf,1024,sizeof(char)) >0 ) { 
		 if(buf[0]=='s') { 
			// обнаружено препятствие, экстренное торможение
			fprintf(stdout,"[i] barrier detected\n");
			barrier=1;
		 }
	  }
	  */
   }

   return 0;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// найден объект
int bot_obj_found(){
   lost_obj = FRAMES_LOST_OBJ ; // сбрасываем счётчик потери объекта
   if(barrier) { 
	  //bot_obj_tracker(); // препятствие
   } else { 
	  bot_obj_pathfinder(); // едем за объектом
   }
   return 0;
}
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// объект потерян
int bot_obj_lost(){
   // если объект не обнаружен в течении FRAMES_LOST_OBJ кадров ...
   if( lost_obj-- > 0 ) { return 1; }  

   // ... то останавливаемся 
   snprintf(cmd,CMD_LEN,"%c",CMD_MOT_STOP); 
   port_ctrl_cmd(cmd,strnlen(cmd,CMD_LEN)); 

   return 0;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// едем за объектом
int bot_obj_pathfinder(){
   int ang;
   
   // корректируем горизонтальное положение бота 
   if( obj_shift_x != 0 ) {
	  // угол поворота
	  //ang=round(((double)abs(obj_shift_x)/(frame->width*0.5))*10.0) ;
	  ang=1;
      snprintf(cmd,CMD_LEN,"%c%i%c", (obj_shift_x>0) ? CMD_MOT_LEFT : CMD_MOT_RIGHT,ang,CMD_MOT_EXE ); 
	  port_ctrl_cmd(cmd,strnlen(cmd,CMD_LEN));
	  fprintf( stdout, "[i] %s\n", ( cmd[0]==CMD_MOT_LEFT )?"left":"right" ); 
   } else {
	  // snprintf(cmd,CMD_LEN,"%c", CMD_MOT_FORWARD ); 
	  // port_ctrl_cmd(cmd,strnlen(cmd,CMD_LEN)); 
	  // fprintf(stdout,"[i] forward\n"); 
   }
   return 0;
}

