/************************************************************
* $Id: frame.c,v 1.4 2014/11/21 10:52:02 mechanoid Exp $ 
*       
*       OS : GNU/Linux 3.14.19 
* COMPILER : GNU project C and C++ Compiler  v. 4.9.1
*
*   AUTHOR : Evgeny S. Borisov
*
*  Glushkov Institute of Cybernetics
*  National Academy of Sciences of Ukraine
* 
*    http://www.mechanoid.kiev.ua
*  e-mail : cv@mechanoid.kiev.ua
* 
************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>

#include "video.h"
#include "frame.h"

// минимальный размер контура для детектора 
#define MIN_CONTOUR 8

// размер контрольного окна в % от размера картинки
#define CHECK_AREA_SIZE 20 

int found=0; // объект обнаружен

CvPoint obj_pos; // координаты центра объекта на картинке
float radius; // радиус области с объектом на картинке

CvPoint p_cent ; // центр картинки
float c_radius ; // размер контрольного окна в пикселях

IplImage *frame_c3 ;
IplImage *frame_c1[3];
IplImage *mask[3];

void* jpeg;
size_t jpeg_size=0;

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER; 


CvCapture* capture;
CvMemStorage* storage;

IplConvKernel* kern;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
int frame_done(){ 
   int i;

   cvReleaseMemStorage(&storage);
   cvReleaseStructuringElement( &kern );

   cvReleaseImage(&frame_c3);
   for(i=0;i<3;i++) cvReleaseImage(&frame_c1[i]);
   for(i=0;i<3;i++) cvReleaseImage(&mask[i]);
   free(jpeg);

   video_done();

   return 0;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
int frame_init(){ 
   int r=3;
   int i;

   if( video_init() ) { return 1; }

   storage = cvCreateMemStorage(0);
   if(!storage){ fprintf(stderr,"[E] cvCreateMemStorage\n"); return 2; }

   // kern = cvCreateStructuringElementEx(r*2+1, r*2+1, r, r, CV_SHAPE_ELLIPSE,NULL);
   kern = cvCreateStructuringElementEx(r*2+1, r*2+1, r, r, CV_SHAPE_RECT,NULL);
   if(!kern){ fprintf( stderr, "[E] cvCreateStructuringElementEx\n"); return 3; }

   frame_c3=cvCloneImage(frame);
   if(!frame_c3){ fprintf(stderr,"[E] cvCloneImage\n"); return 4; }

   // отдельные каналы
   for(i=0;i<3;i++) {
	frame_c1[i]=cvCreateImage( cvGetSize(frame), frame->depth, 1 );
	if(!frame_c1[i]){ fprintf(stderr,"[E] cvCreateImage\n"); return 1; }
   }

   // маска 
   for(i=0;i<3;i++) {
	  mask[i]=cvCreateImage( cvGetSize(frame), frame->depth, 1 );
	  if(!mask[i]){ fprintf(stderr,"[E] cvCreateImage\n"); return 1; }
   }

   jpeg=malloc(DATA_BLOCK_SIZE);
   if(!jpeg) { fprintf(stderr,"[E] malloc: %s\n",strerror(errno)); return 2;   }


   // центр картинки
   p_cent.x=(int)frame->width/2 ;
   p_cent.y=(int)frame->height/2 ;
   // размер контрольной области в центре картинки
   c_radius=MAX(frame->width,frame->height)/100.0*CHECK_AREA_SIZE;

   return 0;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// рисуем край контрольной области
int frame_draw_sight() {
   //cvCircle( frame, p_cent, 5.0, CV_RGB(255,0,0), 1, 8, 0 );
   cvCircle( frame, p_cent, c_radius, CV_RGB(255,0,0), 1, 8, 0 );
   return 0;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
int frame_proc(int* o_sx,int* o_sy,int* o_r){ 
   int x,y;

   // получить кадр 
   if(video_get_frame()) { return 1 ; }


   *o_sx=0; *o_sy=0; *o_r=0;

   obj_pos.x=-1; obj_pos.y=-1; radius=0;

   cvCopy(frame,frame_c3,NULL);

   frame_normalize(frame_c3,frame_c3);
   frame_color_filter(frame_c3,frame_c3); // выделяем нужный цвет
   frame_edge(frame_c3,frame_c1[0]); // выделяем края
   frame_contour(frame_c1[0]); // выделение контуров и поиск нужного контура
   frame_draw_sight();
   
   //cvCopy(frame_c3,frame,NULL); // рисуем рабочий буфер

   if(found) { 
	  cvCircle( frame, obj_pos , 2, CV_RGB(0,255,0), 2, 8, 0 ); 
	  cvCircle( frame, obj_pos , radius, CV_RGB(255,0,0), 2, 8, 0 ); 

	  // вычисляем смещение объекта
	  x=p_cent.x-obj_pos.x;
	  y=p_cent.y-obj_pos.y;
	  if(abs(x)>c_radius || abs(y)>c_radius ) {
		 // смещение объекта выходит за рамки контрольном области
		 // нужна коррекция положения
		 *o_sx=x; *o_sy=y;
	  }
	  *o_r=radius;
   }

   frame_encode_jpeg();

   return 0;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
int frame_normalize(IplImage* src,IplImage* dst) {

   cvSplit( src, frame_c1[0], frame_c1[1], frame_c1[2], NULL );
   cvEqualizeHist(frame_c1[0],frame_c1[0]);
   cvEqualizeHist(frame_c1[1],frame_c1[1]);
   cvEqualizeHist(frame_c1[2],frame_c1[2]);
   cvMerge( frame_c1[0], frame_c1[1], frame_c1[2], NULL, dst );

   return 0;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
int frame_color_filter(IplImage *src,IplImage *dst) {
   // ночь HSV
   int t_max[3]={ 180, 160, 256 }; // ограничение сверху
   int t_min[3]={ 155,  80, 200 }; // ограничение снизу

   // карта цветов
   cvSmooth(src,dst,CV_GAUSSIAN,0,0,5,0);  

   cvCvtColor( dst, dst, CV_BGR2HSV ); // конвертируем карту цветов в HSV

   cvSplit( dst, frame_c1[0], frame_c1[1], frame_c1[2], NULL ); // разрезаем карту цветов на каналы 
  
   // фильтруем фон по порогам
   cvThreshold( frame_c1[0], mask[0], (double)t_min[0], 1.0, CV_THRESH_BINARY ); 
   cvThreshold( frame_c1[0], mask[1], (double)t_max[0], 1.0, CV_THRESH_BINARY_INV ); 
   cvMul(mask[0],mask[1],frame_c1[0],1);

   cvThreshold( frame_c1[1], mask[0], (double)t_min[1], 1.0, CV_THRESH_BINARY ); 
   cvThreshold( frame_c1[1], mask[1], (double)t_max[1], 1.0, CV_THRESH_BINARY_INV ); 
   cvMul(mask[0],mask[1],frame_c1[1],1);

   cvThreshold( frame_c1[2], mask[0], (double)t_min[2], 1.0, CV_THRESH_BINARY ); 
   cvThreshold( frame_c1[2], mask[1], (double)t_max[2], 1.0, CV_THRESH_BINARY_INV ); 
   cvMul(mask[0],mask[1],frame_c1[2],1);

   // объединяем маски трёх каналов
   cvMul(frame_c1[0],frame_c1[1],mask[2],1);
   cvMul(frame_c1[2],mask[2], mask[2],1);

   // инвертируем маску
   cvNot(mask[2],mask[2]);
   cvSubS(mask[2],cvScalarAll(254),mask[2],NULL);

   // формируем картинку результата
   cvCvtColor( dst, dst, CV_HSV2BGR ); 
   cvCopy(src,dst,0) ; 
   cvSet(dst,CV_RGB(0,0,0),mask[2]) ;

   return 0; 
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
int frame_edge(IplImage *src,IplImage *dst){ 
   cvCvtColor( src,dst, CV_BGR2GRAY );
   cvThreshold(dst,dst, 1.0, 256.0, CV_THRESH_BINARY ); 
   cvMorphologyEx( dst, dst, NULL, kern, CV_MOP_OPEN , 1 );
   cvCanny(dst,dst, 100,200,3);

   return 0;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
int frame_contour(IplImage* src) {
   CvSeq* contours=NULL;
   CvSeq* h_next=0;
   CvPoint2D32f p;
   CvSeq* c;

   found=0;

   cvClearMemStorage(storage);
   cvFindContours(src,storage,&contours,sizeof(CvContour),CV_RETR_LIST,CV_CHAIN_APPROX_SIMPLE,cvPoint(0,0) );

   if(!contours) { return 2;}

   // поиск контура максимального размера 
   for( c=contours; c!=NULL; c=c->h_next ) {
	  if(c!=contours) { if (h_next->total>=c->total) { h_next->h_next=h_next->h_next->h_next; continue; } }
	  h_next=c;
   }

   if( h_next->total< MIN_CONTOUR ) { return 1; } // нет объекта - слишком маленькие контуры

   found=1;

   //cvDrawContours( frame_c3, h_next, CV_RGB(255,0,255), CV_RGB(128,128,0),2, 2, CV_AA, cvPoint(0,0) ) ;
   cvMinEnclosingCircle(h_next,&p,&radius) ; // минимальная окружность

   obj_pos.x=p.x;
   obj_pos.y=p.y;

   return 0;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
int frame_encode_jpeg(){
   CvMat* mat;
   size_t s;

   pthread_mutex_lock(&mutex);
	  jpeg_size=0;

	  mat = cvEncodeImage(".jpg", frame,0);
	  if(!mat) { fprintf(stderr,"[E] cvEncodeImage\n") ;  return 1; }

	  s=mat->cols*CV_ELEM_SIZE(CV_MAT_TYPE( mat->type))*mat->rows*sizeof(uchar);
	  if(DATA_BLOCK_SIZE<s) { fprintf(stderr,"[E] data buffer to small\n") ; return 1; } 
	  memcpy(jpeg,(void *)mat->data.ptr,s);
	  jpeg_size=s;

	  cvReleaseMat( &mat ) ;
   pthread_mutex_unlock(&mutex);

   return 0;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
void* frame_get_jpeg(size_t* s) {

   pthread_mutex_lock(&mutex);
	  *s=jpeg_size;
   pthread_mutex_unlock(&mutex);
 
   return jpeg;
}

