/************************************************************
* $Id: http_server.c,v 1.6 2014/11/21 10:52:02 mechanoid Exp $ 
*       
*       OS : GNU/Linux 3.14.19 
* COMPILER : GNU project C and C++ Compiler  v. 4.9.1
*
*   AUTHOR : Evgeny S. Borisov
*
*  Glushkov Institute of Cybernetics
*  National Academy of Sciences of Ukraine
* 
*    http://www.mechanoid.kiev.ua
*  e-mail : cv@mechanoid.kiev.ua
* 
************************************************************/

#include <fcntl.h>
#include <sys/stat.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "frame.h"
#include "http_server.h"

#define PORT 8888

#define BASE_DIR "_www"
#define DEFAULT_INDEX "index.html"



struct MHD_Daemon *httpd;

int http_send_error(struct MHD_Connection *connection);
int http_send_stream(struct MHD_Connection *connection);
int http_send_file(struct MHD_Connection *connection,char path[]);

static int
http_answer(void *cls, struct MHD_Connection *connection,
            const char *url, const char *method,
            const char *version, const char *upload_data,
            size_t *upload_data_size, void **con_cls);

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
int http_server_init(){
   httpd = MHD_start_daemon(MHD_USE_THREAD_PER_CONNECTION, PORT, NULL, NULL, &http_answer, NULL, MHD_OPTION_END);
   if(!httpd) { 
	  fprintf(stderr,"[E] http server error\n") ; 
	  return 1;
   }
   fprintf(stdout,"[i] http server listen port %i\n",PORT) ; 
   return 0;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
void http_server_done(){
   MHD_stop_daemon (httpd);
   fprintf(stdout,"[i] close http server\n") ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
static int
http_answer(void *cls, struct MHD_Connection *connection,
            const char *url, const char *method,
            const char *version, const char *upload_data,
            size_t *upload_data_size, void **con_cls)
{

   char path[255];
   int ret;
   const char* action;

   //printf ("New %s request for %s using version %s\n", method, url, version);

   if(strcmp (method, "GET")) { return MHD_NO; }

   action = MHD_lookup_connection_value(connection, MHD_GET_ARGUMENT_KIND, "action");

   // printf("action:%s\n\n",action);
   if( action!=NULL ) {
	  ret=http_send_stream(connection);
   } else if(!strcmp (url, "/")) {
	  snprintf(path,255,"%s/%s",BASE_DIR,DEFAULT_INDEX);
	  ret=http_send_file(connection,path);
   } else {
	  snprintf(path,255,"%s/%s",BASE_DIR,url);
	  ret=http_send_file(connection,path);
   }

   return ret;

}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
static ssize_t data_generator(void *cls, uint64_t pos, char *buf, size_t max) {
   char bon[]="\r\n--BOUNDARYDONOTCROSS\r\n";
   int res,l,b;
   void* buf0;
   size_t buf0_size;

   buf0=frame_get_jpeg(&buf0_size);
   if(buf0_size<1) { fprintf(stderr,"[W] empty JPEG frame\n"); return 0; }
   
   snprintf(buf,DATA_BLOCK_SIZE,"Content-Type:image/jpeg\r\n"
								"Content-length: %i\r\n\r\n",buf0_size);
   l=strnlen((char*)buf,max);
   b=strlen(bon);

   if(l+buf0_size+b > max) { 
	  fprintf(stderr,"[E] max size %i of buffer less then size of frame, header and footer %i \n",(int)max,buf0_size+l+b ); 
	  res=0;
   } else {
	  memcpy(buf+l,buf0,buf0_size);
	  memcpy(buf+l+buf0_size,bon,b*sizeof(char));
	  //fprintf(stdout,"[i] write pack, size %i + %i + %i\n",l,buf0_size,b);
	  res=buf0_size+l+b;
   }

   return res;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
int http_send_stream(struct MHD_Connection *connection){
   struct MHD_Response *response;
   int ret;

   /* 
   char buf[]="\r\n--BOUNDARYDONOTCROSS\r\n";
   char buf[]="--BOUNDARYDONOTCROSS";
   response = MHD_create_response_from_buffer(strlen(buf),buf, MHD_RESPMEM_PERSISTENT);
   if( response == NULL ) {
	  fprintf(stderr,"MHD_create_response_from_buffer: %s\n",strerror(errno));
	  return MHD_NO;
   }
   */

   response = MHD_create_response_from_callback(MHD_SIZE_UNKNOWN, DATA_BLOCK_SIZE, &data_generator, NULL, NULL);
   if(NULL==response) {
	  fprintf(stderr,"[E] MHD_create_response_from_callback: %s\n",strerror(errno));
	  http_send_error(connection);
	  return MHD_NO;
   }


   MHD_add_response_header(response,  "Connection", "close");
   MHD_add_response_header(response,  "Server", "MJPG-Streamer/0.2");
   MHD_add_response_header(response,  "Cache-Control", "no-store, no-cache, must-revalidate, pre-check=0, post-check=0, max-age=0");
   MHD_add_response_header(response,  "Pragma", "no-cache");
   MHD_add_response_header(response,  "Expires", "Mon, 3 Jan 2000 12:34:56 GMT");
   MHD_add_response_header(response,  "Content-Type", "multipart/x-mixed-replace;boundary=BOUNDARYDONOTCROSS");

   ret = MHD_queue_response (connection, MHD_HTTP_OK, response);
   MHD_destroy_response (response);

   fprintf(stdout,"[i] send video stream\n");

   return ret;
}
 


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
int http_send_file(struct MHD_Connection *connection,char path[]) {
   struct MHD_Response *response;
   struct stat sbuf;
   int fd;
   int ret;

   fd=open (path, O_RDONLY);
   if(fd<0){ 
	  fprintf(stderr,"[E] open %s : %s\n",path,strerror(errno));
	  http_send_error(connection);
	  return MHD_NO;
   }
   if(fstat (fd, &sbuf)) { 
	  fprintf(stderr,"[E] fstat %s : %s\n",path,strerror(errno));
	  http_send_error(connection);
	  close(fd);
	  return MHD_NO;
   }

   fprintf(stdout,"[i] %s : %i B \n",path,(int)sbuf.st_size);

   response = MHD_create_response_from_fd(sbuf.st_size, fd);
   // response = MHD_create_response_from_fd_at_offset(sbuf.st_size, fd, 0);
   if(NULL==response) {
	  fprintf(stderr,"[E] MHD_create_response_from_fd_at_offset: %s\n",strerror(errno));
	  http_send_error(connection);
	  close(fd);
	  return MHD_NO;
   }

   /*
   MHD_add_response_header(response,  "Connection", "close");
   MHD_add_response_header(response,  "Server", "MJPG-Streamer/0.2");
   MHD_add_response_header(response,  "Cache-Control", "no-store, no-cache, must-revalidate, pre-check=0, post-check=0, max-age=0");
   MHD_add_response_header(response,  "Pragma", "no-cache");
   MHD_add_response_header(response,  "Expires", "Mon, 3 Jan 2000 12:34:56 GMT");
   MHD_add_response_header(response,  "Content-Type", "multipart/x-mixed-replace;boundary=BOUNDARYDONOTCROSS");
   */

   ret = MHD_queue_response (connection, MHD_HTTP_OK, response);
   MHD_destroy_response (response);

   return ret;

}
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
int http_send_error(struct MHD_Connection *connection) {
   struct MHD_Response *response;
   const char errorstr[255] = "<html><body>An internal server error has occured!</body></html>";

   response = MHD_create_response_from_buffer(strnlen (errorstr,255), (void *) errorstr, MHD_RESPMEM_PERSISTENT);
   if(NULL != response) {
	  MHD_queue_response (connection, MHD_HTTP_INTERNAL_SERVER_ERROR, response);
	  MHD_destroy_response(response);
	  return 1;
   } 
   return 0;
}



/*
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
static ssize_t data_generator(void *cls, uint64_t pos, char *buf, size_t max) {
   void* buf0;
   int buf0_size;
   int res,l,b;
   char bon[]="\r\n--BOUNDARYDONOTCROSS\r\n";

   if( (buf0=video_get_jpeg(&buf0_size))==NULL ) { fprintf(stderr,"zero size frame\n"); return 0; }

   snprintf(buf,DATA_BLOCK_SIZE,"Content-Type:image/jpeg\r\n" "Content-length: %i\r\n\r\n",buf0_size);
   // snprintf(buf,DATA_BLOCK_SIZE,"Content-Type:image/jpeg\r\n");
   l=strnlen((char*)buf,max);
   b=strlen(bon);

   if(l+buf0_size+b > max) { 
	  fprintf(stderr,"[W] max size %i of buffer less then size of frame, header and footer %i \n",(int)max,buf0_size+l+b ); 
	  res=0;
   } else {
	  memcpy(buf+l,buf0,buf0_size);
	  memcpy(buf+l+buf0_size,bon,b);
	  //fprintf(stderr,"write pack, size %i + %i + %i\n",l,buf0_size,b);
	  res=buf0_size+l+b;
	  //usleep(500);
   }
   
   free(buf0);

   return res;
   
}
*/



/*
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
static ssize_t data_generator(void *cls, uint64_t pos, char *buf, size_t max) {
   char bon[]="\r\n--BOUNDARYDONOTCROSS\r\n";
   void* buf0[DATA_BLOCK_SIZE];
   int buf0_size;
   int res,l,b;

   if(video_get_jpeg(buf0,DATA_BLOCK_SIZE,&buf0_size)) { fprintf(stderr,"[E] video_get_jpeg\n"); return 0; }
   
   snprintf(buf,DATA_BLOCK_SIZE,"Content-Type:image/jpeg\r\n"
								"Content-length: %i\r\n\r\n",buf0_size);
   l=strnlen((char*)buf,max);
   b=strlen(bon);

   if(l+buf0_size+b > max) { 
	  fprintf(stderr,"[E] max size %i of buffer less then size of frame, header and footer %i \n",(int)max,buf0_size+l+b ); 
	  res=0;
   } else {
	  memcpy(buf+l,buf0,buf0_size);
	  memcpy(buf+l+buf0_size,bon,b*sizeof(char));
	  //fprintf(stdout,"[i] write pack, size %i + %i + %i\n",l,buf0_size,b);
	  res=buf0_size+l+b;
   }

   return res;
}

*/


