/************************************************************
* $Id: port_ctrl.c,v 1.5 2014/11/21 11:07:14 mechanoid Exp mechanoid $
*       
*       OS : GNU/Linux 3.14.19 
* COMPILER : GNU project C and C++ Compiler  v. 4.9.1
*
*   AUTHOR : Evgeny S. Borisov
*
*  Glushkov Institute of Cybernetics
*  National Academy of Sciences of Ukraine
* 
*    http://www.mechanoid.kiev.ua
*  e-mail : cv@mechanoid.kiev.ua
* 
************************************************************/

#include <stdio.h>   
#include <termios.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>   
#include <sys/ioctl.h>
#include <stdint.h>

#include "port_ctrl.h"

#define SRV_PORT "/dev/ttyS1"
// #define SRV_PORT "/dev/ttyACM0"
// #define SRV_PORT "/dev/ttyAMA0"


int port_ctrl_getchar();
int port_ctrl_data_avail();


int port_ctrl_fd=0; // порт управления приводами и др.

// - - - - - - - - - - - - - - - - - - - - - - - - -
int port_ctrl_init() {
   struct termios options;
   int status;

   port_ctrl_fd = open (SRV_PORT, O_RDWR | O_NOCTTY | O_NDELAY | O_NONBLOCK);
   if( port_ctrl_fd == -1 ) {
	  fprintf(stderr,"[E] open serial port: %s\n",strerror(errno)); 
	  return 1; 
   }
   
   if( fcntl(port_ctrl_fd, F_SETFL, O_RDWR) == -1 ) {
	  fprintf(stderr,"[E] fcntl: %s\n",strerror(errno)); 
	  return 4; 
   }
  
   if (tcgetattr(port_ctrl_fd, &options) == -1) {
	  fprintf(stderr,"[E] tcgetattr: %s\n",strerror(errno)); 
	  return 2;
   }

   cfmakeraw   (&options) ;
   cfsetispeed (&options, B9600) ;
   cfsetospeed (&options, B9600) ;

   options.c_cflag |= (CLOCAL | CREAD) ;
   options.c_cflag &= ~PARENB ;
   options.c_cflag &= ~CSTOPB ;
   options.c_cflag &= ~CSIZE ;
   options.c_cflag |= CS8 ;
   options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG) ;
   options.c_oflag &= ~OPOST ;

   options.c_cc [VMIN]  =   0 ;
   options.c_cc [VTIME] = 10 ;	// 1 seconds (10 deciseconds)

   if( tcsetattr(port_ctrl_fd, TCSANOW| TCSAFLUSH, &options) == -1 ) {
	  fprintf(stderr,"[E] tcsetattr: %s\n",strerror(errno)); 
	  return 3;
   }

   if( ioctl (port_ctrl_fd, TIOCMGET, &status) == -1 ) {
	  fprintf(stderr,"[E] ioctl: %s\n",strerror(errno)); 
	  return 5;
   }

   status |= TIOCM_DTR ;
   status |= TIOCM_RTS ;

   if( ioctl (port_ctrl_fd, TIOCMGET, &status) == -1 ) {
	  fprintf(stderr,"[E] ioctl: %s\n",strerror(errno)); 
	  return 6;
   }

   usleep (10000) ;	// 10mS

   fprintf(stdout,"[i] port ctrl activated\n") ;

   return 0;
}

// - - - - - - - - - - - - - - - - - - - - - - - - -
int port_ctrl_done() {
   close(port_ctrl_fd);
   return 0;
}


// - - - - - - - - - - - - - - - - - - - - - - - - -
int port_ctrl_cmd(char* cmd,int size) {
   if( write(port_ctrl_fd,cmd,size) <0 ) {
	  fprintf(stderr,"[E] write serial port: %s\n",strerror(errno)); 
	  return 1; 
   }
   return 0;
}

// - - - - - - - - - - - - - - - - - - - - - - - - -
int port_ctrl_answer(char* buf,int len,int size) {
   int c,i;

   for( i=0 ; i<len && port_ctrl_data_avail() ; i++ ) {
	  c = port_ctrl_getchar();
	  if(c<0) { return -1; }
	  buf[i]=(char)c;
   }
   buf[i]=0;
   return i;
}

// - - - - - - - - - - - - - - - - - - - - - - - - -
int port_ctrl_data_avail() {
   int status ;
   if( ioctl (port_ctrl_fd, FIONREAD, &status) == -1 ) {
	  fprintf(stderr,"[E] ioctl: %s\n",strerror(errno)); 
	  return 0;
   }
   return status ;
}

// - - - - - - - - - - - - - - - - - - - - - - - - -
int port_ctrl_getchar() {
  uint8_t x ;

  if(read(port_ctrl_fd, &x, 1) != 1){
	  fprintf(stderr,"[E] read: %s\n",strerror(errno)); 
	  return -1;
   }
   return ((int)x) & 0xFF ;
}


// void serialFlush (const int fd) { tcflush (fd, TCIOFLUSH) ; }
